# -*- coding: utf-8 -*-

__copyright__    = 'Copyright (C) 2023 Masaaki Uematsu'
__version__      = 'ver_0.3.4'
__license__      = 'CC BY-NC-SA 4.0'
__author__       = 'Masaaki Uematsu'
__email__        = 'mu84@cornell.edu'
__url__          = 'https://github.com/MasaakiU/MultiplexNanopore'
__appname__ = "SAVEMONEY"
__comment__ = "written by MU"
__description__ = "Simple Algorithm for Very Efficient Multiplexing of Oxford Nanopore Experiments for You!"

__all__ = [
    "__author__", 
    "__copyright__", 
    "__email__", 
    "__license__", 
    "__appname__", 
    "__url__", 
    "__version__", 
    "__description__", 
    "__comment__", 
]

